#####################################################################
#####################################################################
## Copyright 2005 Steven R. Newcomb                                ##
##                                                                 ##
## This software is licensed under the Apache License, Version 2.0 ##
## (the "License"); you may not use this file except in compliance ##
## with the License.  Copies of the License are available at       ##
##        http://www.apache.org/licenses/LICENSE-2.0               ##
##                                                                 ##
## Unless required by applicable law or agreed to in writing,      ##
## software distributed under the License is distributed on an "AS ##
## IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either ##
## express or implied.  See the License for the specific language  ##
## governing permissions and limitations under the License.        ##
##                                                                 ##
## Copies of the source code and the license are available at      ##
##        http://www.versavant.org                                 ##
#####################################################################
#####################################################################


#####################################################################
## Implementation of the conferral rule "inferFrom_a_sip_"
## in Topic Map Application "BIGASSERT".
#####################################################################


import sys

def inferFrom_a_sip_( vsv, conferralRule, **args):

##         a_sip properties trigger conferrals as follows:

##            - a t-proxy, with its t_sip and t_assertions properties.
##              The t_sip property's value is a list of references to the
##              set of r-proxies (the ones that are referenced as the
##              first member of each tuple in the a_sip property's
##              value).  The t_assertions property's value is a list of
##              proxy references with one item in it - a reference to
##              the a-proxy that has this a_sip property.

##            - for each tuple in the a_sip property value, the following
##              are conferred into existence:

##               . a c-proxy, with its c_sip property.  (See c_sip,
##                 below.)

##               . an r-proxy, with its r_type and r_castings property,
##                 plus an SIP that is not governed by this TMA but is
##                 already extant.

##            - for each tuple in the a_sip property value that has two
##              members (i.e., that has a reference to an x-proxy)

##               . an x-proxy, with its x_castings property, plus the SIP
##                 not governed by this TMA but already extant

##            - an a_type property on the same proxy in which the a_sip
##              property appears

##            - an a_castings property on the same proxy in which the
##              a_sip appears



    topicMap = conferralRule.topicMap  ## make the below code less verbose (and maybe run slightly faster)
    TMA = conferralRule.TMA  ## this is the TMA that defines this conferral rule
    a_sip_overheadDict = TMA.propertyClasses[ 'a_sip'].overheadDict

    a_sip_hashableValueList = a_sip_overheadDict.keys()

    for a_sip_hashableValue in a_sip_hashableValueList:

        a_proxyIdList = a_sip_overheadDict[ a_sip_hashableValue]

        for a_proxyId in a_proxyIdList:

            a_proxy = topicMap.proxies[ a_proxyId]
            a_sip_instance = a_proxy.propertyInstances[ vsv.catKey( TMA.TMAName, 'a_sip')]

            ## Don't be triggered by an a_sip_instance property more than once
            conferralRuleTriggerRoleId = vsv.catKey( conferralRule.conferralRuleKey, 'a_sip value')
            if a_sip_instance.conferralTriggerAncestry.has_key( conferralRuleTriggerRoleId):
                continue

            triggers = {}  ## new dictionary to work in
            triggers[ 'a_sip value'] = a_sip_instance.propertyInstanceKey
            conferredValueSources = {}  ## new dictionary to work in
            conferredValueSources[ 'a_sip value'] = a_sip_instance.propertyInstanceKey

            if vsv.conferralMayBeSignificant( conferralRule, triggers, conferredValueSources):

                ## new t_proxy
                t_proxy = vsv.newProxyObject( topicMap)

                r_proxyIdList = []
                new_c_proxyIdList = []
                for tuple in a_sip_instance._value:
                    r_proxyIdList.append( tuple[ 0])

                    ## new c_proxy
                    c_proxy = vsv.newProxyObject( topicMap)
                    new_c_proxyIdList.append( c_proxy.proxyId)

                    if len( tuple) == 1:
                        c_sip_value = [ tuple[ 0], a_proxy.proxyId]
                    else:  ## len( tuple) == 2:
                        c_sip_value = [ tuple[ 0], a_proxy.proxyId, tuple[ 1]]

                        ## new x_proxy
                        old_x_proxy = topicMap.proxies[ tuple[ 1]]
                        old_x_proxy_sip = vsv.SIPOfProxy( old_x_proxy)

                        x_proxy = vsv.newProxyObject( topicMap)

                        vsv.newPropertyInstance( x_proxy,  ## copy the r_proxy's SIP
                                                 old_x_proxy_sip.propertyClass.TMA,
                                                 old_x_proxy_sip.propertyClass.propertyClassName,
                                                 old_x_proxy_sip._value,
                                                 triggers,
                                                 conferredValueSources)

                        vsv.newPropertyInstance( x_proxy,
                                                 conferralRule.TMA,
                                                 'x_castings',
                                                 [ c_proxy.proxyId],
                                                 triggers,
                                                 conferredValueSources)



                    vsv.newPropertyInstance( c_proxy,
                                             conferralRule.TMA,
                                             'c_sip',
                                             c_sip_value,
                                             triggers,
                                             conferredValueSources)

                    ## new r_proxy
                    old_r_proxy = topicMap.proxies[ tuple[ 0]]
                    old_r_proxy_sip = vsv.SIPOfProxy( old_r_proxy)

                    r_proxy = vsv.newProxyObject( topicMap)

                    vsv.newPropertyInstance( r_proxy,  ## copy the r_proxy's SIP
                                             old_r_proxy_sip.propertyClass.TMA,
                                             old_r_proxy_sip.propertyClass.propertyClassName,
                                             old_r_proxy_sip._value,
                                             triggers,
                                             conferredValueSources)

                    vsv.newPropertyInstance( r_proxy,
                                             conferralRule.TMA,
                                             'r_type',
                                             [ t_proxy.proxyId],
                                             triggers,
                                             conferredValueSources)

                    vsv.newPropertyInstance( r_proxy,
                                             conferralRule.TMA,
                                             'r_castings',
                                             [ c_proxy.proxyId],
                                             triggers,
                                             conferredValueSources)



                ## set t_proxy's properties
                vsv.newPropertyInstance( t_proxy,
                                         conferralRule.TMA,
                                         't_sip',
                                         r_proxyIdList,
                                         triggers,
                                         conferredValueSources)

                vsv.newPropertyInstance( t_proxy,
                                         conferralRule.TMA,
                                         't_assertions',
                                         [ a_proxyId],
                                         triggers,
                                         conferredValueSources)


                ## set a_proxy's properties
                new_a_proxy = vsv.newProxyObject( topicMap)

                vsv.newPropertyInstance( new_a_proxy,  ## copy the a_proxy's SIP
                                         conferralRule.TMA,
                                         'a_sip',
                                         a_sip_instance._value,
                                         triggers,
                                         conferredValueSources)

                vsv.newPropertyInstance( new_a_proxy,
                                         conferralRule.TMA,
                                         'a_type',
                                         [ t_proxy.proxyId],
                                         triggers,
                                         conferredValueSources)

                vsv.newPropertyInstance( new_a_proxy,
                                         conferralRule.TMA,
                                         'a_castings',
                                         new_c_proxyIdList,
                                         triggers,
                                         conferredValueSources)

